#include "player.h"
#include "camera.h"
#include "conf.h"
#include "input.h"
#include "level.h"
#include "util.h"
#include "vec.h"
#include <gint/display.h>

static struct Player self;

static int collide_solid(int x, int y);

void
player_init(struct Vec pos)
{
	self.pos = pos;
	self.spd = VECFZ;
	self.rem = VECFZ;
}

void
player_update(void)
{
	const int on_ground = collide_solid(self.pos.x, self.pos.y + 1);
	const struct Vec dir = VEC(input_down(K_RIGHT) - input_down(K_LEFT),
	                           input_down(K_DOWN) - input_down(K_UP));

	/* acceleration & friction */
	if (on_ground && !dir.x) {
		self.spd.x *= 1.0f * FRICTION_BREAK;
	} else {
		self.spd.x *=
		    1.0f - (on_ground ? FRICTION_GROUND : FRICTION_AIR);
		self.spd.x +=
		    (on_ground ? ACCELERATION_GROUND : ACCELERATION_AIR) *
		    dir.x;
	}

	/* resistance & gravity */
	self.spd.y *= 1.0f - AIR_RESISTANCE;
	self.spd.y += GRAVITY;

	player_move(player_update_rem());
}

void
player_draw(void)
{
	const struct Vec off = camera_offset();
	const int x1 = self.pos.x + off.x;
	const int y1 = self.pos.y + off.y;
	const int x2 = x1 + PLAYER_WIDTH - 1;
	const int y2 = y1 + PLAYER_HEIGHT - 1;
	drect(self.pos.x, self.pos.y, x2, y2, C_BLUE);
}

struct Vec
player_update_rem(void)
{
	struct VecF spd_n_rem = {self.spd.x + self.rem.x,
	                         self.spd.y + self.rem.y};
	struct Vec spd_trunc = {spd_n_rem.x, spd_n_rem.y};
	self.rem.x = spd_n_rem.x - (float)spd_trunc.x;
	self.rem.y = spd_n_rem.y - (float)spd_trunc.y;
	return spd_trunc;
}

void
player_move(struct Vec spd)
{
	float sign_x = signf(spd.x);
	const float sign_y = signf(spd.y);
	if (!sign_x && !sign_y)
		sign_x = 1.0f;

	self.pos.x += spd.x;
	if (collide_solid(self.pos.x, self.pos.y)) {
		self.spd.x = 0.0f;
		self.rem.x = 0.0f;
	}
	while (collide_solid(self.pos.x, self.pos.y)) {
		self.pos.x -= sign_x;
	}

	self.pos.y += spd.y;
	if (collide_solid(self.pos.x, self.pos.y)) {
		self.spd.y = 0.0f;
		self.rem.y = 0.0f;
	}
	while (collide_solid(self.pos.x, self.pos.y)) {
		self.pos.y -= sign_y;
	}
}

static int
collide_solid(int x, int y)
{
	const int x2 = x + PLAYER_WIDTH - 1;
	const int y2 = y + PLAYER_HEIGHT - 1;
	return level_get_px(x, y) == TILE_SOLID ||
	       level_get_px(x, y2) == TILE_SOLID ||
	       level_get_px(x2, y) == TILE_SOLID ||
	       level_get_px(x2, y2) == TILE_SOLID;
}

struct Vec *
player_pos(void)
{
	return &self.pos;
}
